import lmfit
import numpy as np
from scraps.fitsS21 import utils


def offset(freqs, re0, im0):
    """Complex offset re + j*im.

    Freqs vector is ignored, but required for lmfit Model."""
    return re0 + 1j * im0


def mag(freqs, g0, g1, g2):
    """2nd order polynomial.

    References the freqs-array midpoint, which is important because the baseline coefs shouldn't drift
    around with changes in f0 due to power or temperature."""
    x = utils.reduce_by_midpoint(freqs)
    return g0 + g1 * x + g2 * x ** 2


def phase(freqs, p0, p1, p2):
    """Angle in complex plane parameterized by 2nd order polynomial.

    References the freqs-array midpoint, which is important because the baseline coefs shouldn't drift
    around with changes in f0 due to power or temperature."""
    x = utils.reduce_by_midpoint(freqs)
    phi = p0 + p1 * x + p2 * x ** 2
    return np.exp(1j * phi)


class ModelMagBaseline(lmfit.Model):
    __doc__ = (
        "lmfit model that fits a 2nd order polynomial as a function of reduced frequency."
        + lmfit.models.COMMON_DOC
    )

    def __init__(self, *args, **kwargs):
        super().__init__(mag, *args, **kwargs)

    def guess(self, data, freqs=None, mask=None, **kwargs):
        """Mask is an number for how many points to keep on each end. If int, then assumes
        number of points, if float then assumes percent of data. If None, keeps everything."""

        if freqs is None:
            raise ValueError("Must pass a frequencies vector")

        mag_data = np.abs(data)
        xvals = utils.reduce_by_midpoint(freqs)

        masked_data = utils.mask_array_ends(mag_data, mask)
        masked_xvals = utils.mask_array_ends(xvals, mask)

        g2, g1, g0 = np.polyfit(masked_xvals, masked_data, 2)

        params = self.make_params(g2=g2, g1=g1, g0=g0)

        return lmfit.models.update_param_vals(params, self.prefix, **kwargs)


class ModelPhaseBaseline(lmfit.Model):
    __doc__ = (
        "lmfit model that fits a 2nd order polynomial to phase angle as a function of reduced frequency."
        + lmfit.models.COMMON_DOC
    )

    def __init__(self, *args, **kwargs):
        super().__init__(phase, *args, **kwargs)

    def guess(self, data, freqs=None, mask=None, phase_step=None, **kwargs):
        """Treats phase angle as either a 1st or 2nd order polynomial.

        Passing a float to phase_step will cause that amount to be subtracted
        from the first half of the masked data prior to fitting a polynomial.
        This is useful if it's clear that the phase rolls sharply
        through either pi or 2pi, but it's also obvious there is some strong
        frequency-dependent behavior.

        Passing fit_quadratic_phase will cause the baseline to be modeled as
        a second-order polynomial. The default value for this is False."""

        if freqs is None:
            raise ValueError("Must pass a frequencies vector")

        fit_quadratic_phase = kwargs.pop("fit_quadratic_phase", False)
        unwrap_phase = kwargs.pop("unwrap_phase", True)

        phase_data = np.angle(data)
        xvals = utils.reduce_by_midpoint(freqs)

        if unwrap_phase:
            phase_data = np.unwrap(phase_data)

        masked_data = utils.mask_array_ends(phase_data, mask)
        masked_xvals = utils.mask_array_ends(xvals, mask)

        if phase_step:
            step_vector = np.ones_like(masked_data)
            step_vector[-len(step_vector) // 2 :] = 0
            step_vector *= phase_step
            masked_data -= step_vector

        if fit_quadratic_phase:
            p2, p1, p0 = np.polyfit(masked_xvals, masked_data, 2)
        else:
            p1, p0 = np.polyfit(masked_xvals, masked_data, 1)
            p2 = 0

        params = self.make_params(p2=p2, p1=p1, p0=p0)

        if not fit_quadratic_phase:
            params[f"{self.prefix}p2"].set(vary=False)

        return lmfit.models.update_param_vals(params, self.prefix, **kwargs)


class ModelComplexOffset(lmfit.Model):
    __doc__ = "lmfit model that fits a complex offset." + lmfit.models.COMMON_DOC

    def __init__(self, *args, **kwargs):
        super().__init__(offset, *args, **kwargs)

    def guess(self, data, freqs=None, mask=None, **kwargs):
        """Guesses the real and imaginary offsets as the mean of the
        masked real and imaginary parts of data, respectively."""

        masked_data = utils.mask_array_ends(data, mask)

        re = np.real(masked_data).mean()
        im = np.imag(masked_data).mean()

        params = self.make_params(re=re, im=im)

        return lmfit.models.update_param_vals(params, self.prefix, **kwargs)



from scraps.fitsS21 import (
    cmplxIQ_fit,
    cmplxIQ_params,
    hanger_fit,
    hanger_params,
    inline_fit,
    inline_params,
)
from scraps.fitsSweep import f0_tlsAndMBT, qi_tlsAndMBT
from scraps.plot_tools import (
    plotResListData,
    plotResSweep3D,
    plotResSweepParamsVsPwr,
    plotResSweepParamsVsTemp,
    plotResSweepParamsVsX,
)
from scraps.process_file import process_file
from scraps.resonator import Resonator, indexResList, makeResFromData, makeResList
from scraps.resonator_sweep import ResonatorSweep



from scraps.fitsS21.cmplxIQ import cmplxIQ_fit, cmplxIQ_params
from scraps.fitsS21.hanger_resonator import hanger_fit, hanger_params
from scraps.fitsS21.inline_ground_terminated_resonator_s11 import (
    inline_ground_terminated_fit,
    inline_ground_terminated_params,
)
from scraps.fitsS21.inline_resonator import inline_fit, inline_params



"""Functions for fitting hanger-style resonators.

These are resonators with an inverted Lorentzian lineshape (i.e. 0 on resonance, 1 off resonance).
"""

import warnings

import lmfit as lf
import numpy as np
import scipy.signal as sps
import scipy.special as spc


def cmplx_hanger(freqs, f0, df, qc, q0):
    """Assymetric function that fits a hanger or notch type inverted Lorentzian."""
    ft = f0 + df
    x = (freqs - ft) / ft
    dx = df / ft

    qi = q0 * qc / (qc - q0)

    s21 = (1.0 / qi + 1j * 2.0 * (x + dx)) / (1.0 / q0 + 1j * 2.0 * x)
    return s21


def hanger_fit(paramsVec, res, residual=True, **kwargs):
    """Return complex S21 of hanger resonator model or, if data is specified, a residual.

    Parameters
    ----------
    params : list-like
        A an ``lmfit.Parameters`` object containing (df, f0, qc, qi, gain0, gain1, gain2, pgain0, pgain1, pgain2)
    res : scraps.Resonator object
        A Resonator object.
    residual : bool
        Whether to return a residual (True) or to return the model calcuated at the frequencies present in res (False).

    Keyword Arguments
    -----------------
    freqs : list-like
        A list of frequency points at which to calculate the model. Only used if `residual=False`

    remove_baseline : bool
        Whether or not to remove the baseline during calculation (i.e. ignore pgain and gain polynomials). Default is False.

    only_baseline: bool
        Whether or not to calculate and return only the baseline. Default is False.

    Returns
    -------
    model or (model-data)/eps : ``numpy.array``
        If residual=True is specified, the return is the residuals weighted by the uncertainties. If residual=False, the return is the model
        values calculated at the frequency points. The returned array is in the form
        ``I + Q`` or ``residualI + residualQ``.

    """
    # Check if the paramsVec looks like a lmfit params object. If so, unpack to list
    if hasattr(paramsVec, "valuesdict"):
        paramsDict = paramsVec.valuesdict()
        paramsVec = [value for value in paramsDict.values()]

    # intrinsic resonator parameters
    df = paramsVec[0]  # frequency shift due to mismatched impedances
    f0 = paramsVec[1]  # resonant frequency
    qc = paramsVec[2]  # coupling Q
    qi = paramsVec[3]  # internal Q

    # 0th, 1st, and 2nd terms in a taylor series to handle magnitude gain different than 1
    gain0 = paramsVec[4]
    gain1 = paramsVec[5]
    gain2 = paramsVec[6]

    # 0th and 1st terms in a taylor series to handle phase gain different than 1
    pgain0 = paramsVec[7]
    pgain1 = paramsVec[8]

    # Need to add an if statement here to keep compatibility with older version
    # Will be deprecated in the future
    if len(paramsVec) == 12:
        pgain2 = paramsVec[9]

        # Voltage offset at mixer output. Not needed for VNA
        Ioffset = paramsVec[10]
        Qoffset = paramsVec[11]
    else:
        warnings.warn(
            "Warning: new model also fits for quadratic phase. Setting pgain2 = 0. \
                    \nIf using Resonator.do_lmfit() pass kwarg: pgain2_vary = False for legacy behavior."
        )
        pgain2 = 0

        # Voltage offset at mixer output. Not needed for VNA
        Ioffset = paramsVec[9]
        Qoffset = paramsVec[10]

    if residual == False:
        # Use a custom vector of frequencies
        freqs = np.array(kwargs.pop("freqs", res.freq))
    else:
        # Grab a shortcut to the resonant frequency
        freqs = res.freq

    # Repackage resonator data in 1D vector form
    data = np.concatenate((res.I, res.Q), axis=0)

    if (res.sigmaI is not None) and (res.sigmaQ is not None):
        cmplxSigma = np.concatenate((res.sigmaI, res.sigmaQ), axis=0)
    else:
        cmplxSigma = None

    # Gain should reference the file midpoint
    # This is important because the baseline coefs shouldn't drift
    # around with changes in f0 due to power or temperature

    # Of course, this philosophy goes out the window if different sweeps have
    # different ranges.
    fm = freqs[int(np.round((len(freqs) - 1) / 2.0))]
    ffm = (freqs - fm) / fm

    # Calculate the total Q_0
    q0 = 1.0 / (1.0 / qi + 1.0 / qc)

    # Calculate magnitude and phase gain
    gain = gain0 + gain1 * ffm + 0.5 * gain2 * ffm ** 2
    pgain = np.exp(1j * (pgain0 + pgain1 * ffm + 0.5 * pgain2 * ffm ** 2))

    # Allow for voltage offset of I and Q
    offset = Ioffset + 1j * Qoffset

    remove_baseline = kwargs.pop("remove_baseline", False)

    if remove_baseline:
        # Set the baseline vector to a bunch of ones, with zero imaginary part
        total_gain = np.ones_like(freqs, dtype=np.complex)
    else:
        # Actually calculate the baseline
        total_gain = gain * pgain

    only_baseline = kwargs.pop("only_baseline", False)

    if only_baseline:
        # Only return the baseline
        modelCmplx = total_gain
    else:
        # Calculate model from params at each point in freqs
        modelCmplx = total_gain * cmplx_hanger(freqs, f0, df, qc, q0) + offset

    # Package complex data in 1D vector form
    modelI = np.real(modelCmplx)
    modelQ = np.imag(modelCmplx)
    model = np.concatenate((modelI, modelQ), axis=0)

    # Calculate eps from stdev of first 10 pts of data if not supplied
    if cmplxSigma is None and residual == True:
        epsI = np.std(sps.detrend(res.I[0:10]))
        epsQ = np.std(sps.detrend(res.Q[0:10]))
        cmplxSigma = np.concatenate((np.full_like(res.I, epsI), np.full_like(res.Q, epsQ)))

    # Return model or residual
    if residual == True:
        return (model - data) / cmplxSigma
    else:
        return model


def hanger_params(res, **kwargs):
    """Initialize fitting parameters used by the hanger_fit function.

    Parameters
    ----------
    res : ``scraps.Resonator`` object
        The object you want to calculate parameter guesses for.

    Keyword Arguments
    -----------------
    fit_quadratic_phase : bool
        This determines whether the phase baseline is fit by a line or a
        quadratic function. Default is False for fitting only a line.

    hardware : string {'VNA', 'mixer'}
        This determines whether or not the Ioffset and Qoffset parameters are
        allowed to vary by default.

    use_filter : bool
        Whether or not to use a smoothing filter on the data before calculating
        parameter guesses. This is especially useful for very noisy data where
        the noise spikes might be lower than the resonance minimum.

    filter_win_length : int
        The length of the window used in the Savitsky-Golay filter that smoothes
        the data when ``use_filter == True``. Default is ``0.1 * len(data)`` or
        3, whichever is larger.

    Returns
    -------
    params : ``lmfit.Parameters`` object

    """

    # Check if some other type of hardware is supplied
    hardware = kwargs.pop("hardware", "VNA")
    assert hardware in ["VNA", "mixer"], "Unknown hardware type! Choose 'mixer' or 'VNA'."

    # Whether or not to use a smoothing filter on the data before making guesses
    use_filter = kwargs.pop("use_filter", False)
    assert use_filter in [True, False], "Must pass boolean to 'use_filter'."

    # Window length of Savitsky-Golay filter (must be odd and >= 3)
    filter_win_length = kwargs.pop("filter_win_length", 0)

    # If no window length is supplied, defult to 1% of the data vector or 3
    if filter_win_length == 0:
        filter_win_length = int(np.round(len(res.mag) / 100.0))
        if filter_win_length % 2 == 0:
            filter_win_length += 1
        if filter_win_length < 3:
            filter_win_length = 3

    assert (filter_win_length % 2 == 1) and (
        filter_win_length >= 3
    ), "Filter window length must be odd and greater than 3."

    # Whether to fit a line or a parabola to the phase:
    fit_quadratic_phase = kwargs.pop("fit_quadratic_phase", False)

    # There shouldn't be any more kwargs left
    if kwargs:
        raise Exception("Unknown keyword argument supplied")

    if use_filter:
        resMag = sps.savgol_filter(res.mag, filter_win_length, 1)
        resPhase = sps.savgol_filter(res.phase, filter_win_length, 1)
        resUPhase = np.unwrap(resPhase)
    else:
        resMag = res.mag
        resPhase = res.phase
        resUPhase = res.uphase

    # Get index of last datapoint
    findex_end = len(res.freq) - 1

    # Set up lmfit parameters object for fitting later

    # Detrend the mag and phase using first and last 5% of data
    findex_5pc = int(len(res.freq) * 0.05)

    findex_center = int(np.round(findex_end / 2))
    f_midpoint = res.freq[findex_center]

    # Set up a unitless, reduced, mipoint frequency for baselines
    ffm = lambda fx: (fx - f_midpoint) / f_midpoint

    magEnds = np.concatenate((resMag[0:findex_5pc], resMag[-findex_5pc:-1]))
    freqEnds = ffm(np.concatenate((res.freq[0:findex_5pc], res.freq[-findex_5pc:-1])))

    # This fits a second order polynomial
    magBaseCoefs = np.polyfit(freqEnds, magEnds, 2)

    magBase = np.poly1d(magBaseCoefs)

    # Put this back in the resonator object because it is super useful!
    res.magBaseline = magBase(ffm(res.freq))

    # Store the frequency at the magnitude minimum for future use.
    # Pull out the baseline variation first
    findex_min = np.argmin(resMag / magBase(ffm(res.freq)))

    f_at_mag_min = res.freq[findex_min]

    # These points are useful for later code, so add them to the resonator object
    res.fmin = f_at_mag_min
    res.argfmin = findex_min

    # Update best guess with minimum
    f0_guess = f_at_mag_min

    # Update: now calculating against file midpoint
    # This makes sense because you don't want the baseline changing
    # as f0 shifts around with temperature and power

    # Remove any linear variation from the phase (caused by electrical delay)
    phaseEnds = np.concatenate((resUPhase[0:findex_5pc], resUPhase[-findex_5pc:-1]))

    if fit_quadratic_phase:
        phase_poly_order = 2
    else:
        phase_poly_order = 1

    # This fits a second order polynomial
    phaseBaseCoefs = np.polyfit(freqEnds, phaseEnds, phase_poly_order)
    phaseBase = np.poly1d(phaseBaseCoefs)

    # Add to resonator object
    res.phaseBaseline = phaseBase(ffm(res.freq))

    # Set some bounds (resonant frequency should not be within 5% of file end)
    f_min = res.freq[findex_5pc]
    f_max = res.freq[findex_end - findex_5pc]

    if f_min < f0_guess < f_max:
        pass
    else:
        f0_guess = res.freq[findex_center]

    # Guess the Q values:
    # 1/Q0 = 1/Qc + 1/Qi
    # Q0 = f0/fwhm bandwidth
    # Q0/Qi = min(mag)/max(mag)
    magMax = res.magBaseline[findex_min]
    magMin = resMag[findex_min]

    fwhm = np.sqrt((magMax ** 2 + magMin ** 2) / 2.0)
    fwhm_mask = resMag < fwhm
    bandwidth = res.freq[fwhm_mask][-1] - res.freq[fwhm_mask][0]
    q0_guess = f0_guess / bandwidth

    qi_guess = q0_guess * magMax / magMin

    qc_guess = 1.0 / (1.0 / q0_guess - 1.0 / qi_guess)

    # Create a lmfit parameters dictionary for later fitting
    # Set up assymetric lorentzian parameters (Name, starting value, range, vary, etc):
    params = lf.Parameters()
    params.add("df", value=0, vary=True)
    params.add("f0", value=f0_guess, min=f_min, max=f_max, vary=True)
    params.add("qc", value=qc_guess, min=1, max=10 ** 8, vary=True)
    params.add("qi", value=qi_guess, min=1, max=10 ** 8, vary=True)

    # Allow for quadratic gain variation
    params.add("gain0", value=magBaseCoefs[2], min=0, vary=True)
    params.add("gain1", value=magBaseCoefs[1], vary=True)
    params.add("gain2", value=magBaseCoefs[0], vary=True)

    # Allow for linear phase variation
    params.add("pgain0", value=phaseBaseCoefs[phase_poly_order], vary=True)
    params.add("pgain1", value=phaseBaseCoefs[phase_poly_order - 1], vary=True)

    if fit_quadratic_phase:
        params.add("pgain2", value=phaseBaseCoefs[0], vary=True)
    else:
        params.add("pgain2", value=0, vary=False)

    # Add in complex offset (should not be necessary on a VNA, but might be needed for a mixer)
    if hardware == "VNA":
        params.add("Ioffset", value=0, vary=False)
        params.add("Qoffset", value=0, vary=False)
    elif hardware == "mixer":
        params.add("Ioffset", value=0, vary=True)
        params.add("Qoffset", value=0, vary=True)

    return params



"""Functions for fitting ground-terminated inline/transmission resonators measured in reflection.

These are resonators with a traditional Lorentzian lineshape (i.e. 1 on resonance, 0 off resonance).
"""

import types
import warnings

import lmfit
import numpy as np
from scraps.fitsS21 import baselines, utils


def cmplx_grounded_inline_s11(freqs, f0, df, qc, q0):
    """An asymmetric lineshape vs frequency.

    Pass df = 0 to recover the standard complex Lorentzian shape."""

    ft = f0 + df
    x = (freqs - ft) / ft
    dx = df / ft

    s11 = (1 - 2 * (q0 / qc) + 2j * q0 * (x + 2 * dx)) / (1 + 2j * q0 * x)

    return s11


class ModelInlineGroundedResonatorS11(lmfit.model.Model):
    __doc__ = (
        "lmfit model that fits the complex transmission (S_21) of an inline/transmission style resonator."
        + lmfit.models.COMMON_DOC
    )

    def __init__(self, *args, **kwargs):
        super().__init__(cmplx_grounded_inline_s11, *args, **kwargs)

        self.set_param_hint("qc", min=0)
        self.set_param_hint("q0", min=0)

    def guess(self, data, freqs=None, **kwargs):
        """Guess some reasonable initial values for fit parameters.

        The built-in phase guesser does not work well here. Recommend passing fit_baseline="mag",
        and then explicitly setting which phase params you want to vary by hand. Can override any of
        the guessed values by passing param_name=value as keyword argument."""

        if freqs is None:
            raise ValueError("Must pass a frequencies vector")

        # Calculate magnitude and phase
        mag_s11 = np.abs(data)
        uphase_s11 = np.unwrap(np.angle(data))

        # Resonant frequency is probably where the peak is
        min_s11 = mag_s11.min()
        max_s11 = mag_s11.max()
        argmin_s11 = np.argmin(mag_s11)
        f0 = freqs[argmin_s11]

        # Calculate the linewidth to guess the Q
        fwhm_height = (max_s11 ** 2 - min_s11 ** 2) / 2
        fwhm_mask = mag_s11 ** 2 < (fwhm_height + min_s11 ** 2)
        fwhm_left = freqs[fwhm_mask].min()
        fwhm_right = freqs[fwhm_mask].max()
        bandwidth = fwhm_right - fwhm_left

        q0 = f0 / bandwidth

        # A reasonable minimum bound for Q0 is center frequency divided by
        # the bandwidth of the full dataset
        q0_min = f0 / (freqs.max() - freqs.min())

        # A reasonable maximum bound is the center frequency divided by
        # the minimum frequency spacing of the dataset
        q0_max = f0 / np.abs(np.ma.masked_equal(np.diff(freqs), 0)).min()

        if not q0_max > q0 > q0_min:
            q0 = np.sqrt(q0_min * q0_max)
            warnings.warn(
                "q0 = f0/fwhm_bandwith results in impossible value."
                + " Falling back on sqrt(q0_min*q0_max)."
            )

        # The magnitude is symmetric under exchange of qi and qc
        # Can use the phase to lift the degeneracy
        qi_qc_ratios = [
            np.tan(uphase_s11[np.where(freqs == fwhm_right)])[0],
            -np.tan(uphase_s11[np.where(freqs == fwhm_left)])[0],
        ]

        if all(qi_qc_ratios >= 1):
            qc = 2 * q0 / (1 + min_s11)
        elif all(qi_qc_ratios <= 1):
            qc = 2 * q0 / (1 - min_s11)
        else:
            warnings.warn(
                "Too much assymetry to guess qc well. May need to swap qc and qi to get a good fit."
            )
            qc = 2 * q0 / (1 + min_s11)

        # TODO: Guess this properly using phase info
        df = 0

        params = self.make_params(f0=f0, df=df, qc=qc, q0=q0)
        params[f"{self.prefix}q0"].set(min=q0_min, max=q0_max)

        # If allowing df to vary, then it's actually the sum f0+df that
        # tracks the resonant peak
        params.add(f"{self.prefix}ft", expr="f0+df", min=freqs.min(), max=freqs.max())

        return lmfit.models.update_param_vals(params, self.prefix, **kwargs)


# Make up a composite model for an inline resonator with arbitrary baseline and offset
complex_baseline = baselines.ModelMagBaseline() * baselines.ModelPhaseBaseline()
inline_resonator_ground_terminated_full = (
    ModelInlineGroundedResonatorS11() * complex_baseline + baselines.ModelComplexOffset()
)


def inline_resonator_ground_terminated_full_guess(self, data, freqs, mask=0.05, **kwargs):

    mag_mask = kwargs.pop("mag_mask", mask)
    phase_mask = kwargs.pop("phase_mask", mask)
    offset_mask = kwargs.pop("offset_maks", mask)

    fit_baseline = kwargs.pop("fit_baseline", False)
    fit_offset = kwargs.pop("fit_offset", False)

    use_filter = kwargs.pop("use_filter", False)

    if use_filter:
        re_filt = utils.filter_data(np.real(data))
        im_filt = utils.filter_data(np.imag(data))
        data = re_filt + 1j * im_filt

    fit_baseline_vals = [True, False, "mag", "phase", "both"]

    if fit_baseline not in fit_baseline_vals:
        raise ValueError(
            f"Invalid value for fit_baseline. Allowed values are {', '.join(fit_baseline_vals)}."
        )

    if fit_offset not in [True, False]:
        raise ValueError("Invalid value for fit_offset. Allowed values are True or False.")

    # Grab best guess for baseline params
    # Left = inline resonator
    # Right = mag gain * phase gain

    if fit_baseline in [True, "mag", "both"]:
        gain_params = self.left.right.left.guess(data, freqs, mag_mask, **kwargs)
    else:
        gain_params = self.left.right.left.make_params(g0=1, g1=0, g2=0)
        gain_params[f"{self.prefix}g0"].set(vary=False)
        gain_params[f"{self.prefix}g1"].set(vary=False)
        gain_params[f"{self.prefix}g2"].set(vary=False)

    if fit_baseline in [True, "phase", "both"]:
        phase_params = self.left.right.right.guess(data, freqs, phase_mask, **kwargs)
    else:
        phase_params = self.left.right.right.make_params(p0=0, p1=0, p2=0)
        phase_params[f"{self.prefix}p0"].set(vary=False)
        phase_params[f"{self.prefix}p1"].set(vary=False)
        phase_params[f"{self.prefix}p2"].set(vary=False)

    if fit_offset:
        offset_params = self.right.guess(data, freqs, offset_mask, **kwargs)
    else:
        offset_params = self.right.make_params(re=0, im=0)
        offset_params[f"{self.prefix}re0"].set(vary=False)
        offset_params[f"{self.prefix}im0"].set(vary=False)

    # Calculate the baselines
    mag_baseline_guess = self.left.right.left.eval(gain_params, freqs=freqs)
    phase_baseline_guess = self.left.right.right.eval(phase_params, freqs=freqs)
    offset_guess = self.right.eval(offset_params, freqs=freqs)

    # Try and make a best guess for clean data
    reduced_data = (data - offset_guess) / (mag_baseline_guess * phase_baseline_guess)

    # Calculate baseline params from best-guess clean data
    inline_params = self.left.left.guess(reduced_data, freqs, **kwargs)

    # Final params is all of the params concatenated
    params = inline_params + gain_params + phase_params + offset_params

    return lmfit.models.update_param_vals(params, "", **kwargs)


# Attach this to the model instance so it behaves like the non-composite models
inline_resonator_ground_terminated_full.guess = types.MethodType(
    inline_resonator_ground_terminated_full_guess, inline_resonator_ground_terminated_full
)


def inline_ground_terminated_fit(paramsVec, res, residual=True, **kwargs):
    """Wrapper function to make compatible with scraps interface.

    Will deprecate in next major version of scraps."""

    s21_concat = np.concatenate([res.I, res.Q], axis=0)
    freqs = res.freq

    if res.sigmaI is not None and res.sigmaQ is not None:
        sigma_i = res.sigmaI
        sigma_q = res.sigmaQ
    else:
        sigma_i = np.ones_like(freqs)
        sigma_q = np.ones_like(freqs)

    sigma_concat = np.concatenate([sigma_i, sigma_q], axis=0)

    model = inline_resonator_ground_terminated_full.eval(params=paramsVec, freqs=freqs, **kwargs)

    model_concat = np.concatenate([np.real(model), np.imag(model)], axis=0)

    if residual:
        return (model_concat - s21_concat) / sigma_concat
    else:
        return model_concat


def inline_ground_terminated_params(res, **kwargs):
    """Wrapper function to make compatible with scraps interface.

    Will deprecate in next major version of scraps."""
    s21 = res.I + 1j * res.Q
    freqs = res.freq

    return inline_resonator_ground_terminated_full.guess(data=s21, freqs=freqs, **kwargs)



"""A set of simple utility functions for array math."""
import numpy as np
import scipy.signal as sps


def reduce_by_midpoint(array):
    """Subtract off and divide by middle array element.

    Sorts the array before picking mid-point, but returned
    array is not sorted."""
    midpoint = sorted(array)[int(np.round((len(array) - 1) / 2.0))]
    return (array - midpoint) / midpoint


def filter_data(array, filter_win_length=0):
    """Filter data with a Savitsky-Golay filter of window-length
    filter_win_length.

    filter_win_length must be odd and >= 3. This function will
    enforce that requirement by adding 1 to filter_win_length
    until it is satisfied."""

    # If no window length is supplied, defult to 1% of the data vector or 3
    if filter_win_length == 0:
        filter_win_length = int(np.round(len(array) / 100.0))
        if filter_win_length % 2 == 0:
            filter_win_length += 1
        if filter_win_length < 3:
            filter_win_length = 3

    return sps.savgol_filter(array, filter_win_length, 1)


def mask_array_ends(array, mask=None):
    """Return the ends of an array.

    If mask is an int, returns mask items from each end of array.
    If mask is a float, treats mask as a fraction of array length.
    If mask is a an array or a slice, return array[mask]."""
    if mask is None:
        masked_array = array
    elif type(mask) == float:
        pct_mask = int(len(array) * mask)
        masked_array = np.concatenate((array[:pct_mask], array[-pct_mask:]))
    elif type(mask) == int:
        masked_array = np.concatenate((array[:mask], array[-mask:]))
    elif type(mask) in [np.array, list, slice]:
        masked_array = array[mask]
    else:
        raise ValueError("Mask type must be number, array, or slice")

    return masked_array



import warnings

from scraps.fitsS21 import hanger_resonator

warnings.warn(
    DeprecationWarning(
        "This module has been deprecated in favor of scraps.fitsS21.hanger_resonator"
    )
)


def cmplxIQ_fit(paramsVec, res, residual=True, **kwargs):
    """Return complex S21 resonance model or, if data is specified, a residual.

    This function is deprecated and will be removed in a future version. Use hanger_resonator.hanger_fit.

    Parameters
    ----------
    params : list-like
        A an ``lmfit.Parameters`` object containing (df, f0, qc, qi, gain0, gain1, gain2, pgain0, pgain1, pgain2)
    res : scraps.Resonator object
        A Resonator object.
    residual : bool
        Whether to return a residual (True) or to return the model calcuated at the frequencies present in res (False).

    Keyword Arguments
    -----------------
    freqs : list-like
        A list of frequency points at which to calculate the model. Only used if `residual=False`

    remove_baseline : bool
        Whether or not to remove the baseline during calculation (i.e. ignore pgain and gain polynomials). Default is False.

    only_baseline: bool
        Whether or not to calculate and return only the baseline. Default is False.

    Returns
    -------
    model or (model-data)/eps : ``numpy.array``
        If residual=True is specified, the return is the residuals weighted by the uncertainties. If residual=False, the return is the model
        values calculated at the frequency points. The returned array is in the form
        ``I + Q`` or ``residualI + residualQ``.

    """

    warnings.warn(
        DeprecationWarning(
            "This function has been renamed hanger_resonator.hanger_fit. cmplxIQ_fit will be removed in a future version"
        )
    )

    return hanger_resonator.hanger_fit(paramsVec, res, residual, **kwargs)


def cmplxIQ_params(res, **kwargs):
    """Initialize fitting parameters used by the cmplxIQ_fit function.

    Parameters
    ----------
    res : ``scraps.Resonator`` object
        The object you want to calculate parameter guesses for.

    Keyword Arguments
    -----------------
    fit_quadratic_phase : bool
        This determines whether the phase baseline is fit by a line or a
        quadratic function. Default is False for fitting only a line.

    hardware : string {'VNA', 'mixer'}
        This determines whether or not the Ioffset and Qoffset parameters are
        allowed to vary by default.

    use_filter : bool
        Whether or not to use a smoothing filter on the data before calculating
        parameter guesses. This is especially useful for very noisy data where
        the noise spikes might be lower than the resonance minimum.

    filter_win_length : int
        The length of the window used in the Savitsky-Golay filter that smoothes
        the data when ``use_filter == True``. Default is ``0.1 * len(data)`` or
        3, whichever is larger.

    Returns
    -------
    params : ``lmfit.Parameters`` object

    """

    warnings.warn(
        DeprecationWarning(
            "This function has been renamed hanger_resonator.hanger_params. cmplxIQ_fit will be removed in a future version"
        )
    )

    return hanger_resonator.hanger_params(res, **kwargs)
